class Rect {

    static collideRectRect(x, y, w, h, x2, y2, w2, h2) {
        //2d
        //add in a thing to detect rectMode CENTER
        if (x + w >= x2 &&    // r1 right edge past r2 left
            x <= x2 + w2 &&    // r1 left edge past r2 right
            y + h >= y2 &&    // r1 top edge past r2 bottom
            y <= y2 + h2) {    // r1 bottom edge past r2 top
            return true;
        }
        return false;
    };

    static collideRectCircle(rx, ry, rw, rh, cx, cy, diameter) {
        //2d
        // temporary variables to set edges for testing
        const testX = cx;
        const testY = cy;

        // which edge is closest?
        if (cx < rx) {
            testX = rx       // left edge
        } else if (cx > rx + rw) { testX = rx + rw }   // right edge

        if (cy < ry) {
            testY = ry       // top edge
        } else if (cy > ry + rh) { testY = ry + rh }   // bottom edge

        // // get distance from closest edges
        const distance = Rect.dist(cx, cy, testX, testY)

        // if the distance is less than the radius, collision!
        if (distance <= diameter / 2) {
            return true;
        }
        return false;
    };

    static collideCircleCircle(x, y, d, x2, y2, d2) {
        //2d
        if (Rect.dist(x, y, x2, y2) <= (d / 2) + (d2 / 2)) {
            return true;
        }
        return false;
    };

    static collidePointCircle(x, y, cx, cy, d) {
        //2d
        if (Rect.dist(x, y, cx, cy) <= d / 2) {
            return true;
        }
        return false;
    };

    static collidePointEllipse(x, y, cx, cy, dx, dy) {
        //2d
        const rx = dx / 2;
        const ry = dy / 2;
        // Discarding the points outside the bounding box
        if (x > cx + rx || x < cx - rx || y > cy + ry || y < cy - ry) {
            return false;
        }
        // Compare the point to its equivalent on the ellipse
        const xx = x - cx;
        const yy = y - cy;
        const eyy = ry * Math.sqrt(Math.abs(rx * rx - xx * xx)) / rx;
        return yy <= eyy && yy >= -eyy;
    };

    static collidePointRect(pointX, pointY, x, y, xW, yW) {
        //2d
        if (pointX >= x &&         // right of the left edge AND
            pointX <= x + xW &&    // left of the right edge AND
            pointY >= y &&         // below the top AND
            pointY <= y + yW) {    // above the bottom
            return true;
        }
        return false;
    };

    static collidePointLine(px, py, x1, y1, x2, y2, buffer) {
        // get distance from the point to the two ends of the line
        const d1 = Rect.dist(px, py, x1, y1);
        const d2 = Rect.dist(px, py, x2, y2);

        // get the length of the line
        const lineLen = Rect.dist(x1, y1, x2, y2);

        // since floats are so minutely accurate, add a little buffer zone that will give collision
        if (buffer === undefined) { buffer = 0.1; }   // higher # = less accurate

        // if the two distances are equal to the line's length, the point is on the line!
        // note we use the buffer here to give a range, rather than one #
        if (d1 + d2 >= lineLen - buffer && d1 + d2 <= lineLen + buffer) {
            return true;
        }
        return false;
    }

    static collideLineCircle(x1, y1, x2, y2, cx, cy, diameter) {
        // is either end INSIDE the circle?
        // if so, return true immediately
        const inside1 = this.collidePointCircle(x1, y1, cx, cy, diameter);
        const inside2 = this.collidePointCircle(x2, y2, cx, cy, diameter);
        if (inside1 || inside2) return true;

        // get length of the line
        const distX = x1 - x2;
        const distY = y1 - y2;
        const len = Math.sqrt((distX * distX) + (distY * distY));

        // get dot product of the line and circle
        const dot = (((cx - x1) * (x2 - x1)) + ((cy - y1) * (y2 - y1))) / this.pow(len, 2);

        // find the closest point on the line
        const closestX = x1 + (dot * (x2 - x1));
        const closestY = y1 + (dot * (y2 - y1));

        // is this point actually on the line segment?
        // if so keep going, but if not, return false
        const onSegment = this.collidePointLine(closestX, closestY, x1, y1, x2, y2);
        if (!onSegment) return false;

        // draw a debug circle at the closest point on the line
        if (this._collideDebug) {
            this.ellipse(closestX, closestY, 10, 10);
        }

        // get distance to closest point
        distX = closestX - cx;
        distY = closestY - cy;
        const distance = Math.sqrt((distX * distX) + (distY * distY));

        if (distance <= diameter / 2) {
            return true;
        }
        return false;
    }

    static collideLineLine(x1, y1, x2, y2, x3, y3, x4, y4, calcIntersection) {

        let intersection;
        let intersectionX;
        let intersectionY;

        // calculate the distance to intersection point
        const uA = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / ((y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1));
        const uB = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / ((y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1));

        // if uA and uB are between 0-1, lines are colliding
        if (uA >= 0 && uA <= 1 && uB >= 0 && uB <= 1) {

            if (this._collideDebug || calcIntersection) {
                // calc the point where the lines meet
                intersectionX = x1 + (uA * (x2 - x1));
                intersectionY = y1 + (uA * (y2 - y1));
            }

            if (this._collideDebug) {
                this.ellipse(intersectionX, intersectionY, 10, 10);
            }

            if (calcIntersection) {
                intersection = {
                    "x": intersectionX,
                    "y": intersectionY
                }
                return intersection;
            } else {
                return true;
            }
        }
        if (calcIntersection) {
            intersection = {
                "x": false,
                "y": false
            }
            return intersection;
        }
        return false;
    }

    static collideLineRect(x1, y1, x2, y2, rx, ry, rw, rh, calcIntersection) {

        // check if the line has hit any of the rectangle's sides. uses the collideLineLine function above
        let left, right, top, bottom, intersection;

        if (calcIntersection) {
            left = this.collideLineLine(x1, y1, x2, y2, rx, ry, rx, ry + rh, true);
            right = this.collideLineLine(x1, y1, x2, y2, rx + rw, ry, rx + rw, ry + rh, true);
            top = this.collideLineLine(x1, y1, x2, y2, rx, ry, rx + rw, ry, true);
            bottom = this.collideLineLine(x1, y1, x2, y2, rx, ry + rh, rx + rw, ry + rh, true);
            intersection = {
                "left": left,
                "right": right,
                "top": top,
                "bottom": bottom
            }
        } else {
            //return booleans
            left = this.collideLineLine(x1, y1, x2, y2, rx, ry, rx, ry + rh);
            right = this.collideLineLine(x1, y1, x2, y2, rx + rw, ry, rx + rw, ry + rh);
            top = this.collideLineLine(x1, y1, x2, y2, rx, ry, rx + rw, ry);
            bottom = this.collideLineLine(x1, y1, x2, y2, rx, ry + rh, rx + rw, ry + rh);
        }

        // if ANY of the above are true, the line has hit the rectangle
        if (left || right || top || bottom) {
            if (calcIntersection) {
                return intersection;
            }
            return true;
        }
        return false;
    }


    static collidePointPoly(px, py, vertices) {
        let collision = false;

        // go through each of the vertices, plus the next vertex in the list
        let next = 0;
        for (let current = 0; current < vertices.length; current++) {

            // get next vertex in list if we've hit the end, wrap around to 0
            next = current + 1;
            if (next == vertices.length) next = 0;

            // get the PVectors at our current position this makes our if statement a little cleaner
            const vc = vertices[current];    // c for "current"
            const vn = vertices[next];       // n for "next"

            // compare position, flip 'collision' variable back and forth
            if (((vc.y > py && vn.y < py) || (vc.y < py && vn.y > py)) &&
                (px < (vn.x - vc.x) * (py - vc.y) / (vn.y - vc.y) + vc.x)) {
                collision = !collision;
            }
        }
        return collision;
    }

    // POLYGON/CIRCLE
    static collideCirclePoly(cx, cy, diameter, vertices, interior) {

        if (interior == undefined) {
            interior = false;
        }

        // go through each of the vertices, plus the next vertex in the list
        let next = 0;
        for (let current = 0; current < vertices.length; current++) {

            // get next vertex in list if we've hit the end, wrap around to 0
            next = current + 1;
            if (next == vertices.length) next = 0;

            // get the PVectors at our current position this makes our if statement a little cleaner
            const vc = vertices[current];    // c for "current"
            const vn = vertices[next];       // n for "next"

            // check for collision between the circle and a line formed between the two vertices
            const collision = this.collideLineCircle(vc.x, vc.y, vn.x, vn.y, cx, cy, diameter);
            if (collision) return true;
        }

        // test if the center of the circle is inside the polygon
        if (interior == true) {
            const centerInside = this.collidePointPoly(cx, cy, vertices);
            if (centerInside) return true;
        }

        // otherwise, after all that, return false
        return false;
    }

    static collideRectPoly(rx, ry, rw, rh, vertices, interior) {
        if (interior == undefined) {
            interior = false;
        }

        // go through each of the vertices, plus the next vertex in the list
        let next = 0;
        for (let current = 0; current < vertices.length; current++) {

            // get next vertex in list if we've hit the end, wrap around to 0
            next = current + 1;
            if (next == vertices.length) next = 0;

            // get the PVectors at our current position this makes our if statement a little cleaner
            const vc = vertices[current];    // c for "current"
            const vn = vertices[next];       // n for "next"

            // check against all four sides of the rectangle
            const collision = this.collideLineRect(vc.x, vc.y, vn.x, vn.y, rx, ry, rw, rh);
            if (collision) return true;

            // optional: test if the rectangle is INSIDE the polygon note that this iterates all sides of the polygon again, so only use this if you need to
            if (interior == true) {
                const inside = this.collidePointPoly(rx, ry, vertices);
                if (inside) return true;
            }
        }

        return false;
    }

    static collideLinePoly(x1, y1, x2, y2, vertices) {

        // go through each of the vertices, plus the next vertex in the list
        let next = 0;
        for (let current = 0; current < vertices.length; current++) {

            // get next vertex in list if we've hit the end, wrap around to 0
            next = current + 1;
            if (next == vertices.length) next = 0;

            // get the PVectors at our current position extract X/Y coordinates from each
            const x3 = vertices[current].x;
            const y3 = vertices[current].y;
            const x4 = vertices[next].x;
            const y4 = vertices[next].y;

            // do a Line/Line comparison if true, return 'true' immediately and stop testing (faster)
            const hit = this.collideLineLine(x1, y1, x2, y2, x3, y3, x4, y4);
            if (hit) {
                return true;
            }
        }
        // never got a hit
        return false;
    }

    static collidePolyPoly(p1, p2, interior) {
        if (interior == undefined) {
            interior = false;
        }

        // go through each of the vertices, plus the next vertex in the list
        let next = 0;
        for (let current = 0; current < p1.length; current++) {

            // get next vertex in list, if we've hit the end, wrap around to 0
            next = current + 1;
            if (next == p1.length) next = 0;

            // get the PVectors at our current position this makes our if statement a little cleaner
            const vc = p1[current];    // c for "current"
            const vn = p1[next];       // n for "next"

            //use these two points (a line) to compare to the other polygon's vertices using polyLine()
            const collision = this.collideLinePoly(vc.x, vc.y, vn.x, vn.y, p2);
            if (collision) return true;

            //check if the 2nd polygon is INSIDE the first
            if (interior == true) {
                collision = this.collidePointPoly(p2[0].x, p2[0].y, p1);
                if (collision) return true;
            }
        }

        return false;
    }

    static collidePointTriangle(px, py, x1, y1, x2, y2, x3, y3) {

        // get the area of the triangle
        const areaOrig = Math.abs((x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1));

        // get the area of 3 triangles made between the point and the corners of the triangle
        const area1 = Math.abs((x1 - px) * (y2 - py) - (x2 - px) * (y1 - py));
        const area2 = Math.abs((x2 - px) * (y3 - py) - (x3 - px) * (y2 - py));
        const area3 = Math.abs((x3 - px) * (y1 - py) - (x1 - px) * (y3 - py));

        // if the sum of the three areas equals the original, we're inside the triangle!
        if (area1 + area2 + area3 == areaOrig) {
            return true;
        }
        return false;
    }

    static collidePointPoint(x, y, x2, y2, buffer) {
        if (buffer == undefined) {
            buffer = 0;
        }

        if (Rect.dist(x, y, x2, y2) <= buffer) {
            return true;
        }

        return false;
    };

    static collidePointArc(px, py, ax, ay, arcRadius, arcHeading, arcAngle, buffer) {

        if (buffer == undefined) {
            buffer = 0;
        }
        // point
        const point = this.createVector(px, py);
        // arc center point
        const arcPos = this.createVector(ax, ay);
        // arc radius vector
        const radius = this.createVector(arcRadius, 0).rotate(arcHeading);

        const pointToArc = point.copy().sub(arcPos);

        if (point.dist(arcPos) <= (arcRadius + buffer)) {
            const dot = radius.dot(pointToArc);
            const angle = radius.angleBetween(pointToArc);
            if (dot > 0 && angle <= arcAngle / 2 && angle >= -arcAngle / 2) {
                return true;
            }
        }
        return false;
    }

    static dist() {
        if (arguments.length === 4) {
            return Rect.hypot(arguments[2] - arguments[0], arguments[3] - arguments[1]);
        } else if (arguments.length === 6) {
            //3D
            return Rect.hypot(
                arguments[3] - arguments[0],
                arguments[4] - arguments[1],
                arguments[5] - arguments[2]
            );
        }
    };

    // Calculate the length of the hypotenuse of a right triangle
    // This won't under- or overflow in intermediate steps
    // https://en.wikipedia.org/wiki/Hypot
    static hypot(x, y, z) {
        // Use the native implementation if it's available
        if (typeof Math.hypot === 'function') {
            return Math.hypot.apply(null, arguments);
        }

        // Otherwise use the V8 implementation
        // https://github.com/v8/v8/blob/8cd3cf297287e581a49e487067f5cbd991b27123/src/js/math.js#L217
        var length = arguments.length;
        var args = [];
        var max = 0;
        for (var i = 0; i < length; i++) {
            var n = arguments[i];
            n = +n;
            if (n === Infinity || n === -Infinity) {
                return Infinity;
            }
            n = Math.abs(n);
            if (n > max) {
                max = n;
            }
            args[i] = n;
        }

        if (max === 0) {
            max = 1;
        }
        var sum = 0;
        var compensation = 0;
        for (var j = 0; j < length; j++) {
            var m = args[j] / max;
            var summand = m * m - compensation;
            var preliminary = sum + summand;
            compensation = preliminary - sum - summand;
            sum = preliminary;
        }
        return Math.sqrt(sum) * max;
    }

    static areaRect(w, h) {
        return w * h;
    }

    static insideRectRect(x, y, w, h, x2, y2, w2, h2) {
        const res = x >= x2 && x + w <= x2 + w2 && y >= y2 && y + h <= y2 + h2;
        return res;
    }

    static equalsRectRect(x, y, w, h, x2, y2, w2, h2, distance = 0) {
        return Math.abs(x - x2) < distance && 
                Math.abs(y - y2) < distance && 
                Math.abs(w - w2) < distance * 2 && 
                Math.abs(h - h2) < distance * 2; 
    }

    static equalsRectObj(rect1, rect2, distance = 0) {
        return Math.abs(rect1.x - rect2.x) < distance && 
                Math.abs(rect1.y - rect2.y) < distance && 
                Math.abs(rect1.w - rect2.w) < distance * 2 && 
                Math.abs(rect1.h - rect2.h) < distance * 2; 
    }

    static intersectRect(x, y, w, h, x2, y2, w2, h2) {

        const collides1 = Rect.collideRectRect(x, y, w, h, x2, y2, w2, h2);
        const collides2 = Rect.collideRectRect(x2, y2, w2, h2, x, y, w, h);
        const inside1 = Rect.insideRectRect(x, y, w, h, x2, y2, w2, h2);
        const inside2 = Rect.insideRectRect(x2, y2, w2, h2, x, y, w, h);
        
        if (!collides1 && !collides2 && !inside1 && !inside2) 
            return null;

        const top = Math.max(y, y2);
        const bottom = Math.min(y + h, y2 + h2);
        const left = Math.max(x, x2);
        const right = Math.min(x + w, x2 + w2);
        return {
            x: left,
            y: top,
            w: right - left,
            h: bottom - top
        }
    }

    static unionRect(x, y, w, h, x2, y2, w2, h2) {

        const top = Math.min(y, y2);
        const bottom = Math.max(y + h, y2 + h2);
        const left = Math.min(x, x2);
        const right = Math.max(x + w, x2 + w2);
        return {
            x: left,
            y: top,
            w: right - left,
            h: bottom - top
        }
    }

    static unionRects(rects) {
        let prev;

        if (!rects.length) return null;
        if (rects.length == 1) return rects[0];

        for (let i = 1; i < rects.length; i++) {
            const r1 = prev ? prev : rects[0];
            const r2 = rects[i];
            prev = Rect.unionRect(r1.x, r1.y, r1.w, r1.h, r2.x, r2.y, r2.w, r2.h);
        }
        
        return prev;
    }

    static inflateRect(x, y, w, h, xscl, yscl = xscl) {
        const xAdj = (w * xscl - w) / 2;
        let yAdj = (h * yscl - h) / 2;

        return {
            x: x - xAdj,
            y: y - yAdj,
            w: w + (xAdj * 2),
            h: h + (yAdj * 2)
            
        }
    }

    static inflateFixedRect(x, y, w, h, xfixed, yfixed = xfixed) {
        return {
            x: x - xfixed,
            y: y - yfixed,
            w: w + (xfixed * 2),
            h: h + (yfixed * 2)
        }
    }

    static insideLineRect = (x1, y1, x2, y2, rx, ry, rw, rh) => {
        const res =
            x1 >= rx &&
            x2 <= rx + rw &&
            y1 >= ry &&
            y2 <= ry + rh
    
        return res;
    }
    
    static intersectLineRect = (x1, y1, x2, y2, rx, ry, rw, rh) => {
        return Rect.collideLineRect(x1, y1, x2, y2, rx, ry, rw, rh) || Rect.insideLineRect(x1, y1, x2, y2, rx, ry, rw, rh);
    }

}

exports = module.exports = Rect;